{
 "cells": [
  {
   "cell_type": "markdown",
   "source": [
    "# Tutorial: Custom gym Environment\r\n",
    "\r\n",
    "In this notebook, an environment is set up with continuous observation_space and discrete action_space.<small>[[source](https://www.youtube.com/watch?v=bD6V3rcr_54)]</small>\r\n"
   ],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "# Importing Dependencies"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 21,
   "source": [
    "# Importing Libraries\r\n",
    "import gym\r\n",
    "from gym import Env\r\n",
    "from gym.spaces import Discrete, Box, Dict, Tuple, MultiBinary, MultiDiscrete \r\n",
    "import numpy as np\r\n",
    "import pandas as pd\r\n",
    "import matplotlib.pyplot as plt\r\n",
    "import random\r\n",
    "import os\r\n",
    "\r\n",
    "from stable_baselines3.common.env_checker import check_env\r\n",
    "from stable_baselines3.common.evaluation import evaluate_policy\r\n",
    "from stable_baselines3 import DQN\r\n",
    "from stable_baselines3 import PPO"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "# Shower Environment"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "source": [
    "class ShowerEnv(Env):\r\n",
    "    def __init__(self):\r\n",
    "        # Actions we can take, down, stay, up\r\n",
    "        self.action_space = Discrete(3)\r\n",
    "        # Temperature array\r\n",
    "        self.observation_space = Box(low=np.array([0]), high=np.array([100]))\r\n",
    "        # Set start temp\r\n",
    "        self.state = 38 + random.randint(-3,3)\r\n",
    "        # Set shower length\r\n",
    "        self.shower_length = 60\r\n",
    "        \r\n",
    "    def step(self, action):\r\n",
    "        # Apply action\r\n",
    "        # 0 -1 = -1 temperature\r\n",
    "        # 1 -1 = 0 \r\n",
    "        # 2 -1 = 1 temperature \r\n",
    "        self.state += action -1 \r\n",
    "        # Reduce shower length by 1 second\r\n",
    "        self.shower_length -= 1 \r\n",
    "        \r\n",
    "        # Calculate reward\r\n",
    "        if self.state >=37 and self.state <=39: \r\n",
    "            reward =1 \r\n",
    "        else: \r\n",
    "            reward = -1 \r\n",
    "        \r\n",
    "        # Check if shower is done\r\n",
    "        if self.shower_length <= 0: \r\n",
    "            done = True\r\n",
    "        else:\r\n",
    "            done = False\r\n",
    "        \r\n",
    "        # Apply temperature noise\r\n",
    "        #self.state += random.randint(-1,1)\r\n",
    "        # Set placeholder for info\r\n",
    "        info = {}\r\n",
    "        \r\n",
    "        # Return step information\r\n",
    "        return self.state, reward, done, info\r\n",
    "\r\n",
    "    def render(self):\r\n",
    "        # Implement viz\r\n",
    "        pass\r\n",
    "    \r\n",
    "    def reset(self):\r\n",
    "        # Reset shower temperature\r\n",
    "        self.state = np.array([38 + random.randint(-3,3)]).astype(float)\r\n",
    "        # Reset shower time\r\n",
    "        self.shower_length = 60 \r\n",
    "        return self.state"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "# Checking Environment"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "source": [
    "# Checking whether our custom environment conforms with the necessary assertions\r\n",
    "env = ShowerEnv()\r\n",
    "check_env(env, warn=False)\r\n",
    "# Observe the state initialization\r\n",
    "env.reset()\r\n",
    "env.render()"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 12,
   "source": [
    "# Explore the state/observation and acion spaces\r\n",
    "print(env.observation_space)\r\n",
    "print(env.action_space)\r\n",
    "\r\n",
    "action_samples = []\r\n",
    "obs_samples = []\r\n",
    "n_samples = 25\r\n",
    "print('\\nNum of Samples: ', n_samples)\r\n",
    "for _ in range(n_samples):\r\n",
    "    action_samples.append(env.action_space.sample())\r\n",
    "    obs_samples.append(env.observation_space.sample())\r\n",
    "print(len(np.unique(action_samples)), ':\\t', np.unique(action_samples))\r\n",
    "print(len(np.unique(obs_samples)), ':\\t', np.unique(obs_samples), '\\n')\r\n",
    "\r\n",
    "# See that box has no attribte 'n'\r\n",
    "print(env.action_space.n)\r\n",
    "print(env.observation_space.n)\r\n",
    "\r\n",
    "del n_samples, action_samples, obs_samples\r\n"
   ],
   "outputs": [
    {
     "output_type": "stream",
     "name": "stdout",
     "text": [
      "Box(0.0, 100.0, (1,), float32)\n",
      "Discrete(3)\n",
      "\n",
      "Num of Samples:  25\n",
      "3 :\t [0 1 2]\n",
      "25 :\t [ 3.5996962  6.4574146  9.578733  13.806476  16.718254  31.083012\n",
      " 32.470306  39.09603   41.21765   41.593693  45.524384  46.594196\n",
      " 51.84048   60.28984   64.59786   69.74104   73.02744   76.58201\n",
      " 78.07482   78.15038   88.742775  92.168594  92.570755  95.42639\n",
      " 97.20068  ] \n",
      "\n",
      "3\n"
     ]
    },
    {
     "output_type": "error",
     "ename": "AttributeError",
     "evalue": "'Box' object has no attribute 'n'",
     "traceback": [
      "\u001b[1;31m---------------------------------------------------------------------------\u001b[0m",
      "\u001b[1;31mAttributeError\u001b[0m                            Traceback (most recent call last)",
      "\u001b[1;32m<ipython-input-12-1ddd63988e80>\u001b[0m in \u001b[0;36m<module>\u001b[1;34m\u001b[0m\n\u001b[0;32m     15\u001b[0m \u001b[1;31m# See that box has no attribte 'n'\u001b[0m\u001b[1;33m\u001b[0m\u001b[1;33m\u001b[0m\u001b[1;33m\u001b[0m\u001b[0m\n\u001b[0;32m     16\u001b[0m \u001b[0mprint\u001b[0m\u001b[1;33m(\u001b[0m\u001b[0menv\u001b[0m\u001b[1;33m.\u001b[0m\u001b[0maction_space\u001b[0m\u001b[1;33m.\u001b[0m\u001b[0mn\u001b[0m\u001b[1;33m)\u001b[0m\u001b[1;33m\u001b[0m\u001b[1;33m\u001b[0m\u001b[0m\n\u001b[1;32m---> 17\u001b[1;33m \u001b[0mprint\u001b[0m\u001b[1;33m(\u001b[0m\u001b[0menv\u001b[0m\u001b[1;33m.\u001b[0m\u001b[0mobservation_space\u001b[0m\u001b[1;33m.\u001b[0m\u001b[0mn\u001b[0m\u001b[1;33m)\u001b[0m\u001b[1;33m\u001b[0m\u001b[1;33m\u001b[0m\u001b[0m\n\u001b[0m\u001b[0;32m     18\u001b[0m \u001b[1;33m\u001b[0m\u001b[0m\n\u001b[0;32m     19\u001b[0m \u001b[1;32mdel\u001b[0m \u001b[0mn_samples\u001b[0m\u001b[1;33m,\u001b[0m \u001b[0maction_samples\u001b[0m\u001b[1;33m,\u001b[0m \u001b[0mobs_samples\u001b[0m\u001b[1;33m\u001b[0m\u001b[1;33m\u001b[0m\u001b[0m\n",
      "\u001b[1;31mAttributeError\u001b[0m: 'Box' object has no attribute 'n'"
     ]
    }
   ],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "# Random action episodes"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 13,
   "source": [
    "# Perform episodes of random actions, i.e. No decision making\r\n",
    "episodes = 5\r\n",
    "for episode in range(1, episodes+1):\r\n",
    "    state = env.reset()\r\n",
    "    done = False\r\n",
    "    score = 0 \r\n",
    "    \r\n",
    "    while not done:\r\n",
    "        env.render()\r\n",
    "        action = env.action_space.sample()\r\n",
    "        n_state, reward, done, info = env.step(action)\r\n",
    "        score+=reward\r\n",
    "    print('Episode:{} Score:{}'.format(episode, score))\r\n",
    "env.close()\r\n",
    "\r\n"
   ],
   "outputs": [
    {
     "output_type": "stream",
     "name": "stdout",
     "text": [
      "Episode:1 Score:-44\n",
      "Episode:2 Score:-54\n",
      "Episode:3 Score:-30\n",
      "Episode:4 Score:-38\n",
      "Episode:5 Score:-30\n"
     ]
    }
   ],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "# Defining DQN model"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 14,
   "source": [
    "model = DQN(\"MlpPolicy\", env, verbose=0)\r\n",
    "model.learn(total_timesteps=100000)"
   ],
   "outputs": [
    {
     "output_type": "execute_result",
     "data": {
      "text/plain": [
       "<stable_baselines3.dqn.dqn.DQN at 0x2837d4df320>"
      ]
     },
     "metadata": {},
     "execution_count": 14
    }
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 15,
   "source": [
    "evaluate_policy(model, env, n_eval_episodes=100, render=False)"
   ],
   "outputs": [
    {
     "output_type": "stream",
     "name": "stderr",
     "text": [
      "C:\\Users\\sarth\\.conda\\envs\\master\\lib\\site-packages\\stable_baselines3\\common\\evaluation.py:69: UserWarning: Evaluation environment is not wrapped with a ``Monitor`` wrapper. This may result in reporting modified episode lengths and rewards, if other wrappers happen to modify these. Consider wrapping environment first with ``Monitor`` wrapper.\n",
      "  UserWarning,\n"
     ]
    },
    {
     "output_type": "execute_result",
     "data": {
      "text/plain": [
       "(-12.0, 58.787753826796276)"
      ]
     },
     "metadata": {},
     "execution_count": 15
    }
   ],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "# Learning model further"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 16,
   "source": [
    "model.learn(total_timesteps=100000)"
   ],
   "outputs": [
    {
     "output_type": "execute_result",
     "data": {
      "text/plain": [
       "<stable_baselines3.dqn.dqn.DQN at 0x2837d4df320>"
      ]
     },
     "metadata": {},
     "execution_count": 16
    }
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 17,
   "source": [
    "evaluate_policy(model, env, n_eval_episodes=100, render=False)"
   ],
   "outputs": [
    {
     "output_type": "execute_result",
     "data": {
      "text/plain": [
       "(-9.6, 59.227020860414704)"
      ]
     },
     "metadata": {},
     "execution_count": 17
    }
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 18,
   "source": [
    "model.learn(total_timesteps=100000)"
   ],
   "outputs": [
    {
     "output_type": "execute_result",
     "data": {
      "text/plain": [
       "<stable_baselines3.dqn.dqn.DQN at 0x2837d4df320>"
      ]
     },
     "metadata": {},
     "execution_count": 18
    }
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 19,
   "source": [
    "evaluate_policy(model, env, n_eval_episodes=100, render=False)"
   ],
   "outputs": [
    {
     "output_type": "execute_result",
     "data": {
      "text/plain": [
       "(-7.2, 59.56643350075611)"
      ]
     },
     "metadata": {},
     "execution_count": 19
    }
   ],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "# Defining PPO model"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 22,
   "source": [
    "model = PPO(\"MlpPolicy\", env, verbose=0)\r\n",
    "model.learn(total_timesteps=100000)"
   ],
   "outputs": [
    {
     "output_type": "execute_result",
     "data": {
      "text/plain": [
       "<stable_baselines3.ppo.ppo.PPO at 0x2837d4f4278>"
      ]
     },
     "metadata": {},
     "execution_count": 22
    }
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 23,
   "source": [
    "evaluate_policy(model, env, n_eval_episodes=100, render=False)"
   ],
   "outputs": [
    {
     "output_type": "execute_result",
     "data": {
      "text/plain": [
       "(59.7, 0.714142842854285)"
      ]
     },
     "metadata": {},
     "execution_count": 23
    }
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 24,
   "source": [
    "model.learn(total_timesteps=100000)"
   ],
   "outputs": [
    {
     "output_type": "execute_result",
     "data": {
      "text/plain": [
       "<stable_baselines3.ppo.ppo.PPO at 0x2837d4f4278>"
      ]
     },
     "metadata": {},
     "execution_count": 24
    }
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 26,
   "source": [
    "evaluate_policy(model, env, n_eval_episodes=100, render=False)"
   ],
   "outputs": [
    {
     "output_type": "execute_result",
     "data": {
      "text/plain": [
       "(59.48, 0.8772684879784526)"
      ]
     },
     "metadata": {},
     "execution_count": 26
    }
   ],
   "metadata": {}
  }
 ],
 "metadata": {
  "orig_nbformat": 4,
  "language_info": {
   "name": "python",
   "version": "3.6.13",
   "mimetype": "text/x-python",
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "pygments_lexer": "ipython3",
   "nbconvert_exporter": "python",
   "file_extension": ".py"
  },
  "kernelspec": {
   "name": "python3",
   "display_name": "Python 3.6.13 64-bit ('master': conda)"
  },
  "interpreter": {
   "hash": "0c640dc386691b1b42b3424c708d4d1eda291d0197a35eddf54ba99eb6d636bd"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}